//
//								SPECTROLOG
// Visualization program for geophysical well data stored in LAS format 
// http://www.epgeology.com/petrophysics-f23/las-file-standards-t2608.html
// Display a selected curve and calculate its spectrogram
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK).
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
// As a general rule, "free software" means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
// Please report all bugs and problems on the following page:
// http://www.uwyo.edu/cmi/
//     
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/Fl_Check_Button.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>
#include <fftw3.h>
/**
 The standard FFTW distribution works most efficiently and without any problem
 for arrays whose size can be factored into small primes (2, 3, 5, and 7).
 It was noticed that for the array lengths with larger-size prime factors (17 and
 above), the complex transforms might occasionally produce incorrect results in
 the form of high-frequency noise.    
 **/
using namespace std;

#define MARGIN		60
#define RIM			12
#define MAX_CURVE	80
#define MAX_BIN	9999	// Maximum number of bins in a histogram
#define FLT_MAX	1E+37
#define FLT_MIN	-1E+37
//
#define Re(z,i) ((z)[2*(i)])
#define Im(z,i) ((z)[2*(i)+1])
//
#define NW		300		// Number of wavelengths to dispaly at log scale
#define DWIN	150.0	// 150.0 meters,  max. window size for spectral analysis
//
#define FOLD	6		// Maximum number of selected spectral lines to stack
//
#define PALSZ	11
//
#ifndef M_PI
#define M_PI 3.14159265359
#endif	// Tells CPP to define M_PI if it isn't defined already
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
FILE *in_fp = NULL;
char inpname[200],c_word[4];
const char *unit[] = {"meter","foot","unknown"};
float no_value=-9999;			// Absent values (gaps in log data) 
int chosen=-1, dep_unit=2;		// Depth unit: 0=meter, 1=foot, 2=unknown
int Ncur=-1;					// Number of curves in file
int Ndat=-1;					// Number of samples in a curve
//
int win_w, win_h, spec_w, spec_h;
//
float depth_sta, depth_end, dS;		// for the whole log length
int samp_sta, samp_fin, samp_vis;	// for selected log interval
float dep_sta, dep_end;				// for displayed part of the log 
float min_val, max_val;				// Min-Max of displayed curves
float *logs, *mini, *maxi;			// Min-Max for all curves
float *avri;
//
char curve[MAX_CURVE][22],cgbuf[40];
int pt_size=2, line_draw=0;
int curve_no = 1;					// default value
Fl_Color bg_color = FL_WHITE;
Fl_Color colr = 0;					// Curve color
/**********************  Spectrogram-Related  *********************/
//
//		WIN - number of data samples corresponding DWIN meters 
int WIN,wls_min,wls_max;
int win_mean;		// Half window size for Mean calculation
float taperk,gain=1.0;
double k_res=1.5;
double *bell[FOLD],norm[FOLD];
//
///////////////////  Color Palettes  ////////////////////
//			array size = PALSZ*4,  palette size = PALSZ
//
float *pal;				// Palette container
//
float BWR_pal[] =  
{  	0, 2.5, 5.0,  10,  25,  35,  60,  70,  80,  90, 100,
	0,   0,   0, 125, 255, 255, 255, 255, 255, 185, 125,
	0,   0, 255, 255, 255, 255, 255, 125,   0,   0,   0,
  125, 255, 255, 255, 255, 125,   0,   0,   0,   0,   0};
//  
float YGB_pal[] = 
{   0, 2.5, 5.0,  10,  25,  35,  60,  70,  80,  90, 100,
  255, 125,   0,   0, 255, 175,  75,   0,   0,   0,   0,
  255, 225, 175, 255, 255, 175,  75,   0,   0,   0,   0,
	0,   0,   0,   0, 255, 255, 255, 255, 205, 160, 120};	
//
float PINK_pal[] = 
{   0, 2.5, 5.0,  10,  25,  35,  60,  70,  80,  90, 100,
  255, 255, 255, 125,   0,   0,   0,   0,   0,   0,   0,
  255, 125,   0,   0,   0, 255, 255, 255, 185, 125,  45,
  255, 255, 255, 255, 255, 255, 125,   0,   0,   0,   0};	
//
float gray_pal[] = 
{   0, 2.5, 5.0,  10,  25,  35,  60,  70,  80,  90, 100,
  255, 225, 200, 175, 150, 127, 105,  80,  55,  25,   0,
  255, 225, 200, 175, 150, 127, 105,  80,  55,  25,   0,
  255, 225, 200, 175, 150, 127, 105,  80,  55,  25,   0};
//
////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/

void read_file_info(const char *name);
float CCI(float *Y, int NS, float X_intr);
void bells(int win_sz, double res_factor);
void pal_interp(float [], int, float, short []);

////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////

Fl_Double_Window *curve_window, *spec_window;
Fl_Window *opts_window;
Fl_Float_Input *min_inp,*max_inp,*dep_sta_inp,*dep_end_inp;
Fl_Button *apply, *specgram, *choose_color, *end_this;
Fl_Value_Slider *balance, *tapersz, *demean, *gainsp;
Fl_Check_Button *connected;

/////////////////////////////////////////////////////////////////////
/****          List of Widgets and Drawing functions            ****/
/////////////////////////////////////////////////////////////////////

/**********     File-choosing Menu    ************/
class File_menu : public Fl_Window {
	Fl_Native_File_Chooser *fc;
	
	void read_data(const char *filename)	// Process the chosen file
	{
		strcpy(inpname,filename);
		chosen=1;			// will wait while(chosen<0)
	}
	
	// Handle an 'Open' request from the menu
	static void open_cb(Fl_Widget *w, void *v) {
		File_menu *choice = (File_menu*)v;
		choice->fc->title("Open");
		choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
		switch ( choice->fc->show() ) {
			case -1: break;	// Error
			case  1: break; // Cancel
			default:		// choice->fc->show() returns 0
				choice->read_data(choice->fc->filename());
				// choice->deactivate();				
				break;
		}
	}
	
	static void quit_cb(Fl_Widget *w, void *v) {
		exit(0);
	}
	
public:
	File_menu() : Fl_Window(500,300,320,160,"Choose the file") {
		Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,300,25);
		menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
		menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
		
		// Describe the chooser
		Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
		box->color(FL_GRAY-1); 
		box->box(FL_FRAME_BOX);
		box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
		box->label("Choose LAS-format file\nwith well-log data");
		box->labelsize(18);
		// Initialize the file chooser
		fc = new Fl_Native_File_Chooser();
		fc->filter("ASCII text files \t*.LAS, *.las\n");			 
		end();
	}
};
/////////////////////////////////////////////////////////////////////
//////							CURVE  DRAWING				/////////
//////					in a double-buffered window			/////////
/////////////////////////////////////////////////////////////////////
class Draw_curve : public Fl_Double_Window 
{
	
public:
  	Draw_curve(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)  // Function draw() within class Draw_seis
	{
		Fl_Double_Window::draw();
		/**    don't declare reserved 'x' and 'y' variables !!!   **/
		int i,j,k;
		int xx, yy, x0, y0, x_dim, y_dim;
		float val,dx,dy,ddy;
		
		x_dim = w() - (MARGIN+RIM);  y_dim = h() - (MARGIN+RIM);	

		fl_color(51);  fl_rectf(0,0,w(),h());			// Gray area
		fl_color(FL_BLACK); fl_rect(0,0,w(),h());		// bounding rect
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(MARGIN,MARGIN,x_dim,y_dim,255,255,255);
		
		//   G L O B A L variables 
		samp_sta = (int)((dep_sta-depth_sta)/dS);
		samp_fin = (int)((dep_end-depth_sta)/dS);
		samp_vis = samp_fin-samp_sta;
		
		dx = (float)x_dim/(max_val-min_val);

		dy = dep_end-dep_sta;
		if(dy>5000.) ddy = 1000.0;
		else if(dy>3000.) ddy = 500.0;
		else if(dy>1500.) ddy = 200.0;
		else if(dy>500.) ddy = 100.0;
		else if(dy>100.) ddy = 50.0;
		else if(dy>50.) ddy = 10.0;
		else if(dy>15.) ddy = 5.0;
		else ddy = 1.0;
		
		//////////////////////////////////////////////////////
		//  Draw tick marks and labels on the Depth axis
		fl_line_style(0,1);	// solid, 1 pixel wide
		fl_font(1,11);			// Helvetica, 11 pixel		
		for(val=dep_sta;val<=dep_end;val+=ddy)
		{
			yy = MARGIN + (int)((val-dep_sta)*y_dim/dy);
			fl_color(23); fl_xyline(MARGIN, yy, w()-RIM);	// gray horiz line
			fl_color(0);  fl_xyline(MARGIN-6, yy, MARGIN);	// horiz line
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,24,yy+4);
		}
		//////////////////////////////////////////////////////
		//  Draw tick marks and labels on the X-axis
		for(i=0;i<5;i++)
		{
			xx = MARGIN + x_dim*i/4;
			yy = MARGIN-20;
			fl_color(23);	fl_yxline(xx, MARGIN, h()-RIM);  // gray vert. line
			fl_color(0);	fl_yxline(xx, MARGIN-6, MARGIN); // vertical line
			val = min_val + (max_val-min_val)*i/4.0;
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,xx-8,MARGIN-12);
		}
		//////////////////////////////////////////////////////
		//  Draw labels on the coordinate axes		
		sprintf(cgbuf,"Depth,  %s",unit[dep_unit]);
		fl_font(1,12);		fl_draw(90,cgbuf,18,y_dim/2);	// rotated 90˚
		sprintf(cgbuf,"%s",curve[curve_no]+3);
		fl_draw(cgbuf,MARGIN+x_dim/3,25);		
		
		//////////////////////////////////////////////////////
		//  Draw Data Points with optional connection lines
		//////////////////////////////////////////////////////
		if(samp_sta<0) x0 = 0; 
		else {	val = logs[Ncur*samp_sta+curve_no];
					x0 = MARGIN + (int)((val - min_val)*dx); }	
		y0 = MARGIN;
		fl_color(colr);		
		for(i=0;i<samp_vis;i++)
		{
			j = samp_sta+i;
			if(j<0) continue;
			if(j>=Ndat) break;
			k = Ncur*j+curve_no;
			val = logs[k];
			if(val<=no_value) continue;	// ?
				
			xx = MARGIN + (int)((val - min_val)*dx);
			if(xx<MARGIN) xx=MARGIN;
			if(xx>=(MARGIN+x_dim)) xx=MARGIN+x_dim;
			yy = MARGIN + i*y_dim/samp_vis;

			if(pt_size>0)
				fl_pie(xx-pt_size/2,yy-pt_size/2,pt_size,pt_size,0.,360.);
			
			if(line_draw==1) fl_line(x0,y0,xx,yy);
			x0 = xx;		y0 = yy;
			
		}
		
		fl_color(FL_BLACK); fl_rect(MARGIN,MARGIN,x_dim,y_dim);		
		
	} 	// void draw(void)
	
} *curve_plot;
//
/////////////////////////////////////////////////////////////////////
//////	 SPECTROGRAM		SPECTROGRAM		 SPECTROGRAM	/////////
//////				in a double-buffered window				/////////
/////////////////////////////////////////////////////////////////////
/**  Spectral lines indexing and their relation to the wavelength:
 INDEX			Wavenumber		WL(wavelength)		  WL
 [feet]		   [samples]		[feet or meters]	[samples]
 0					0			WIN*dS/0			WIN/0
 1				1/(WIN*dS)		WIN*dS/1			WIN/1
 2				2/(WIN*dS)		WIN*dS/2			WIN/2
 ...						...	
 n				n/(WIN*dS)		WIN*dS/n			WIN/n
 ...						...	
 WIN-2			-2/(WIN*dS)		-WIN*dS/2
 WIN-1			-1/(WIN*dS)		-WIN*dS
 ========================================================================
 for(n=WIN/2) wavenumber=1.0/(2.0*dS); wavelength=2.0*dS; (Minimum value)
 Maximum wavelength [in samples] should not exceed WIN samples
 Minimum wavelength should be more than 2 samples
 **/		
class Draw_spec : public Fl_Double_Window 
{
	
public:
  	Draw_spec(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)  // Function draw() within class Draw_seis
	{
		Fl_Double_Window::draw();
		
		int i,ii,j,k,ind,samp,N;
		int xx, yy, x0, y0, x_dim, y_dim;
		int sta,fin,len,Ntap;
		short rgb[3];
		float *orig_data,*log_data,*winbuf;
		float val,xval,dy,ddy,xi,units=1.0;
		double avr,df,max,dval,wl,taper,kx;	
		double *sum,*spec,*tmp;		
		fftw_complex *arr;
		fftw_plan fwd;	
		
		x_dim = w() - (MARGIN+RIM);  y_dim = h() - (MARGIN+RIM);	

		fl_color(51);  fl_rectf(0,0,w(),h());			// Gray area
		fl_color(FL_BLACK); fl_rect(0,0,w(),h());		// bounding rect
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(MARGIN,MARGIN,x_dim,y_dim,255,255,255);
		
		//   G L O B A L variables 
		samp_sta = (int)((dep_sta-depth_sta)/dS);
		samp_fin = (int)((dep_end-depth_sta)/dS);
		samp_vis = samp_fin-samp_sta;
		
		dy = dep_end-dep_sta;
		if(dy>5000.) ddy = 1000.0;
		else if(dy>3000.) ddy = 500.0;
		else if(dy>1500.) ddy = 200.0;
		else if(dy>500.) ddy = 100.0;
		else if(dy>100.) ddy = 50.0;
		else if(dy>50.) ddy = 10.0;
		else if(dy>15.) ddy = 5.0;
		else ddy = 1.0;		

		// Depth unit: 0=meter, 1=foot, 2=unknown
		if(dep_unit==1) units=1.0; else units=3.28;	
		
		/////////////////////////////////////////////////////////////////
		/////////////      Definitions and Memory       /////////////////
	
		N = WIN/2; // Spectral line # that corresponds Nyquist wavenumber

		arr = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * WIN);
		fwd = fftw_plan_dft_1d(WIN,arr,arr,-1,FFTW_ESTIMATE); // in-place
		tmp=(double *)arr; // To translate complex-number type into double

		log_data = (float*) malloc(samp_vis*sizeof(float));
		// orig_data[..] -> to keep track of gaps (no_value) on spectra display
		orig_data = (float*) malloc(samp_vis*sizeof(float)); 	
		sum = (double*)malloc(samp_vis*sizeof(double));				
		winbuf = (float*) malloc(WIN*sizeof(float));
		spec = (double*)malloc(y_dim*N*sizeof(double));
		
		bells(WIN,k_res);
		//for(i=0;i<WIN;i++) cout << i << " " << bell[0][i] << endl;	
	
		///////////////////////////////////////////////////////////////	
		////////    Log-data input, demeaning, and tapeing     ////////		
		///////////////////////////////////////////////////////////////	
		for(i=0;i<samp_vis;i++)	// Data input and moving mean estimate
		{
			val = no_value;			// Initialization
			j = samp_sta+i;
			if(j>=0 && j<Ndat) val = logs[Ncur*j+curve_no];
			orig_data[i] = val;		// 'no_value' if outside or gaps
			
			//  Fill the gaps and outer area with 'global' mean value
			if(val<=no_value) val = avri[curve_no];			
			log_data[i]=val;	// Log data for SPECTROGRAM calculation
			// Incorporate 'global' mean in 'moving' mean calculation
			if(i<=0) sum[i] = val;
			else sum[i] = sum[i-1]+val; // for Mean calculation
		}
		
		// Ntap - tapering length on both sides of the signal
		// taperk - taper coefficient, %	
		Ntap = (int)(taperk*samp_vis/100.0);	j=samp_vis-Ntap;		
		for(i=0;i<samp_vis;i++)		// Data de-meaning and tapeing
		{ 
			taper = 1.0;   // Initialization
			if(i<Ntap) {kx=(double)(i+1-Ntap)/Ntap; taper=fabs(1.0-kx*kx);}
			if(i>j) { kx=(double)(i-j)/Ntap; taper=fabs(1.0-kx*kx); }
		
			sta=i-win_mean; if(sta<0) sta=0;	
			fin=i+win_mean; if(fin>=samp_vis) fin=samp_vis-1;
			len=fin-sta+1;
			avr = (sum[fin]-sum[sta])/(float)len;
			val = log_data[i];
			log_data[i] = (float)(taper*(val-avr));
		}
	
		///////////////////////////////////////////////////////////////	
		///////////////    Calculate SPECTROGRAMM     /////////////////		
		///////////////////////////////////////////////////////////////
		
		for(yy=0; yy<y_dim; yy++)		//   Cycle pixels in Y-direction
		{
			for(i=0;i<N;i++) spec[yy*N+i] = 0.0;		
			k = yy*samp_vis/y_dim;		// log-data index from 0 to samp_vis
			
			for(i=0;i<WIN;i++)
			{
				j = k+i-N;
				if(j>=0 && j<samp_vis) winbuf[i]=log_data[j];
				else winbuf[i] = log_data[k];	// defined from k=0 to samp_vis
			}

			for(j=0;j<FOLD;j++)			// Combined Resolution SPECTROGRAM
			{
				for(i=0;i<WIN;i++)
				{ Im(tmp,i) = 0.0;  Re(tmp,i) = bell[j][i]*winbuf[i]; }
				
				fftw_execute(fwd);
				
				for(i=0;i<N;i++) // Stack spectra for balanced T-F resolution
					spec[yy*N+i]+=sqrt(Re(tmp,i)*Re(tmp,i)+Im(tmp,i)*Im(tmp,i))/norm[j];
			} 		
		}										//  for(depth_pixels)

		max = 0.0;		///////    Search for spectral representative maximum
		for(yy=0;yy<y_dim;yy++)
			for(i=1;i<N;i++) { dval=spec[yy*N+i]; if(dval>max) max=dval; }
		if(max==0.0) max=1.0;	
	
		/////////////////////////////////////////////////////////////////
		/////////////      SPECTRA VISUALIZATION        /////////////////
		
		for(k=0; k<y_dim; k++)				//   Cycle pixels in Y-direction
		{									//		Normalize to abs. maximum
			yy = k*samp_vis/y_dim;			// log-data index from 0 to samp_vis 		
			if(orig_data[yy]<=no_value) continue;	// Gaps in observations	
			for(i=0;i<N;i++) winbuf[i] = (float)(100.0*gain*spec[k*N+i]/max); 
			
			yy = MARGIN + k;				// Y-coordinate for drawings
			
			for(i=0;i<NW;i++)		// NW=300,  Cycle pixels in X-direction
			{	//  resample spectum to a logarithmic scale
				wl = exp2(0.559 + (double)i/35.825)/units;
				// 'wl' range from 1.473 to 479.4 ft
				xi = dS*WIN/wl;
				ind = (int)xi;
				if(ind < 0) continue;				
				if(ind >=N) continue;
				
				val = CCI(winbuf,N,xi);	// Cubic Convolution Interpolation
				if(val>100.0) val=100.0;
				pal_interp(pal,PALSZ,val,rgb);
				xx = MARGIN + i;			// X-coordinate for drawings
				fl_color(rgb[0],rgb[1],rgb[2]); fl_point(xx,yy);
			}			
		}		

		fl_color(0);
		//////////////////////////////////////////////////////
		//  Draw tick marks and labels on the Depth axis
		fl_line_style(0,1);					// solid, 1 pixel wide
		fl_font(1,11);						// Helvetica, 11 pixel		
		for(val=dep_sta;val<=dep_end;val+=ddy)
		{
			yy = MARGIN + (int)((val-dep_sta)*y_dim/dy);
			fl_color(24); fl_xyline(MARGIN, yy, w()-RIM);	// gray horiz line
			fl_color(0);  fl_xyline(MARGIN-6, yy, MARGIN);	// horiz line
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,24,yy+4);
		}
		
		fl_color(0);
		//////////////////////////////////////////////////////
		//  Draw tick marks and labels on the X-axis
		//////////////////////////////////////////////////////
		if(dep_unit==1) xval=1.5; else xval=0.5;
		samp=0;				
		for(i=0;i<NW;i++)		// NW=300,  Cycle pixels in X-direction
		{
			xx = MARGIN + i;
			wl = exp2(0.559 + (double)i/35.825)/units;
			if(wl>=xval)
			{
				if(samp>0) { fl_color(24); fl_yxline(xx, MARGIN, h()-RIM); }
				fl_color(0); fl_yxline(xx, MARGIN-6, MARGIN); // vertical lines
				if(samp>1) { fl_color(0); sprintf(cgbuf,"%3.0f",xval); }
				else sprintf(cgbuf,"%3.1f",xval);
				fl_draw(cgbuf,xx-8,MARGIN-12);
				if(dep_unit==1)		// feet
					switch(samp)
					{
						case 0: xval=3.0; samp++;  break;
						case 1: xval=10.0; samp++;  break;
						case 2: xval=30.0; samp++;  break;						
						case 3: xval=100.0; samp++;  break;					
						case 4: xval=300.0; samp++;  break;
						case 5: samp++;  break;						
						default: break;						
					}
				else
					switch(samp)
					{
						case 0: xval=1.0; samp++;  break;
						case 1: xval=3.0; samp++;  break;
						case 2: xval=10.0; samp++;  break;						
						case 3: xval=30.0; samp++;  break;					
						case 4: xval=100.0; samp++;  break;
						case 5: samp++;  break;						
						default: break;						
					}				
			}
			if(samp>=6) break;
		}

		fl_color(0); sprintf(cgbuf,"Depth,  %s",unit[dep_unit]);
		fl_font(1,12);		fl_draw(90,cgbuf,18,y_dim/2);	// rotated 90˚
		if(dep_unit==1) sprintf(cgbuf,"Wavelength, foot");
		else sprintf(cgbuf,"Wavelength, meters");
		fl_font(1,13); fl_draw(cgbuf,MARGIN+100,25);	
		fl_rect(MARGIN,MARGIN,x_dim,y_dim);
		
		////////    Scalogram-Related  (memory liberation)   //////////
		///////////////////////////////////////////////////////////////
		free(log_data);  free(orig_data);  
		free(sum);  free(winbuf);  free(spec);
		fftw_free(arr);		
		fftw_destroy_plan(fwd);
		///////////////////////////////////////////////////////////////

	} 	// void draw(void)
	
} *spec_plot;
//
//////////////////////////////////////////////////////////////
class Draw_pal : public Fl_Widget  ////  Drawing Color Palette
{
	void draw() 
	{
		int i,j,xx,yy,x_sta,wid,hei;  //val,old,step=2;
		short rgb[3];
		float yi;
		
		x_sta = x();  yy = y();  wid = w();  hei = h();
		fl_rectf(x_sta-2,yy-2,wid+4,hei+4,0,0,0); // Black frame		
		
		for(i=0,xx=x_sta;i<wid;i++,xx++)
		{
			yi=pal[0]+i*(pal[PALSZ-1]-pal[0])/wid;	// actually, yi=i/wid
			pal_interp(pal, PALSZ, yi ,rgb);
			fl_rectf(xx,yy,1,hei,rgb[0], rgb[1], rgb[2]);
		}		
	}      // draw()
	
public:
	Draw_pal(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
	
} *palette;
//
////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////

void input_curve(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_curve=(Fl_Input_Choice *)data;

	char buf[4];
	strncpy(buf,in_curve->value(),3);
	sscanf(buf,"%d", &curve_no);
	
	min_val=mini[curve_no];			 max_val=maxi[curve_no];
 	sprintf(cgbuf," %g",min_val);  min_inp->value(cgbuf);
	sprintf(cgbuf," %g",max_val);  max_inp->value(cgbuf);	
}
void input_min(Fl_Widget *, void *) {
	min_val = atof(min_inp->value());
}
void input_max(Fl_Widget *, void *) {
	max_val = atof(max_inp->value());
}
void input_dep_sta(Fl_Widget *, void *) {
	dep_sta = atof(dep_sta_inp->value());
}
void input_dep_end(Fl_Widget *, void *) {
	dep_end = atof(dep_end_inp->value());
}
//
void spinner_cb(Fl_Widget *w, void *) {
   Fl_Spinner *s = (Fl_Spinner *)w;	
   pt_size = (int)s->value();
}
//
void connect_cb(Fl_Widget *w, void *) {
   Fl_Check_Button *check = (Fl_Check_Button *)w;
   line_draw = check->value();
}
///////    Color Chooser    ////////////
void do_color(Fl_Widget*, void*)
{	
	colr = fl_show_colormap(bg_color);
	//canvas->redraw();
}
/////////////////   SPECTROGRAM   //////////////
//
//////////   Resolution Parameter Slider
static void slider_input(Fl_Widget* balance) 
{
	k_res = ((Fl_Valuator*)balance)->value();	
}
//////////   Taper Length Slider
static void slidek_input(Fl_Widget* tapersz) 
{
	taperk = ((Fl_Valuator*)tapersz)->value();
}
//////////   Gain Slider
static void slideg_input(Fl_Widget* gainsp) 
{
	gain = ((Fl_Valuator*)gainsp)->value();
}
//////////   Window Size Slider
static void slides_input(Fl_Widget* demean) 
{
	float val = ((Fl_Valuator*)demean)->value();
	win_mean = (int)(val/(2.0*dS)); // Half window size, samples
}
//
void choice_pal(Fl_Widget*,void*data) 
{
	Fl_Input_Choice *in_pal=(Fl_Input_Choice *)data;
	
	int num;
	
	strncpy(c_word,in_pal->value(),3);
	sscanf(c_word,"%d", &num);
	switch(num)
	{  
		case 0:	pal = BWR_pal; break;		
		case 1:	pal = YGB_pal; break;
		case 2:	pal = PINK_pal; break;		
		case 3:	pal = gray_pal; break;						
		default: break;
	}	
	palette->redraw();	
}
//
void doSPEC(Fl_Widget *, void*)
{  
	
	// Define Window Size with Margins
	spec_w = MARGIN+RIM+NW;
	spec_h = Fl::h() - 40;
	
	spec_window = new Fl_Double_Window(800,0,spec_w,spec_h,
												"Spectrogram Window");
	spec_plot = new Draw_spec(0,0,spec_w,spec_h);
	
	spec_window->end();
	//curve_window->resizable(seis_window);	
	spec_window->show();
	
}
//
///////////    EXIT    ////////////
void exit_b(Fl_Widget *) 
{
	fclose(in_fp);           exit(0);
}
/////////////////////////////////////////////////////////////////////
///////////////           DRAWING Window          ///////////////////
/////////////////////////////////////////////////////////////////////
void doPLOT(Fl_Widget *, void*)
{  

	// Define Window Size with Margins
	win_w = MARGIN+RIM+200;
	win_h = Fl::h() - 40;

	curve_window = new Fl_Double_Window(600,0,win_w,win_h,
											"Curve Display Window");
	curve_plot = new Draw_curve(0,0,win_w,win_h);
	
	curve_window->end();
	//curve_window->resizable(seis_window);	
	curve_window->show();
	
}
//
/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	int i,cur;
	double t,val;
	const char *unit[] = {"meter","foot"};
	char cbuf1[300],cbuf2[99],cbuf[80];
	///////////////////////////////////////////////////////////
	
	Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
	// NULL use the scheme defined in the FLTK_SCHEME env.var. 
	
	if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
	else	// if you want to create an Apple  *.app bundle application
	{
		File_menu *choice = new File_menu();
		choice->show();
		choice->end(); // no more widgets will be added to the window
		while(chosen<0) Fl::wait();
		choice->hide();
	}
	
	read_file_info(inpname);
	
	opts_window = new Fl_Window(1050,50,420,870,"Display Manager");

	Fl_Box* label1 = new Fl_Box(FL_FRAME_BOX,10,4,400,106,"");
	label1->color(6);
	
	Fl_Box* label2 = new Fl_Box(80,12,250,20,"  File Information  ");
	label2->labelcolor(FL_BLACK);
	label2->labelfont(FL_HELVETICA_BOLD);	label2->labelsize(16);

	sprintf(cbuf1,"Log file contains %d curves,  %d samples each\n\
Starting depth = %.3f     Ending depth = %.3f\n\
Depth increment = %.3f   Unit of measurement = %s\n\
Gaps in observations are filled with   %.3f",
Ncur,Ndat,depth_sta,depth_end,dS,unit[dep_unit],no_value);
	Fl_Box* label3 = new Fl_Box(25,22,360,90,cbuf1);
	label3->labelfont(FL_HELVETICA);	label3->labelsize(14);
	
	//////////////////////////////////////////////////////////////////
	////////////////    L O G    C U R V E      //////////////////////
	//////////////////////////////////////////////////////////////////
	cur=1;
	Fl_Input_Choice in_curve(225,125,176,25,"Select curve by its mnemonic: ");
	in_curve.callback(input_curve, (void*)&in_curve);
	
	for(i=0;i<Ncur;i++) in_curve.add(curve[i]);
	in_curve.value(cur);  // default curve for plotting
	in_curve.labelsize(14); in_curve.labelfont(FL_HELVETICA_BOLD);
	in_curve.align(Fl_Align(FL_ALIGN_LEFT));	
	

	Fl_Box* label4 = new Fl_Box(40,170,110,25,"Depth range to plot:");
	label4->labelfont(FL_HELVETICA_BOLD); label4->labelsize(14);
	
	dep_sta_inp = new Fl_Float_Input(225, 170, 75, 25, " from ");
	sprintf(cbuf," %g",dep_sta);  
	dep_sta_inp->value(cbuf);   dep_sta_inp->callback(input_dep_sta);
	dep_end_inp = new Fl_Float_Input(325, 170, 75, 25, " to ");
	sprintf(cbuf," %g",dep_end);  
	dep_end_inp->value(cbuf);   dep_end_inp->callback(input_dep_end);
	

	Fl_Box* label5 = new Fl_Box(40,210,120,25,"Range of curve values:");
	label5->labelfont(FL_HELVETICA_BOLD); label5->labelsize(14);
	
	min_inp = new Fl_Float_Input(225, 210, 75, 25, " from ");
	sprintf(cbuf," %g",min_val);  
	min_inp->value(cbuf);   min_inp->callback(input_min);
	max_inp = new Fl_Float_Input(325, 210, 75, 25, " to ");
	sprintf(cbuf," %g",max_val);  
	max_inp->value(cbuf);   max_inp->callback(input_max);
	
	Fl_Spinner* spin = new Fl_Spinner(125, 250, 36, 25,
												 "Dot size: ");
	spin->labelsize(14);	spin->labelfont(FL_HELVETICA_BOLD);
	spin->minimum(0);		spin->maximum(7);		spin->step(1);
	spin->value(pt_size);	spin->align(Fl_Align(FL_ALIGN_LEFT));
	spin->callback(spinner_cb);
	Fl_Box* label6 = new Fl_Box(165,250,40,25,"pixel");
	label6->labelfont(FL_HELVETICA); label6->labelsize(14);	
	
	choose_color = new Fl_Button(270,250,100,25,"Dot Color");
	choose_color->labelcolor(FL_BLUE);	choose_color->color(134);
	choose_color->labelfont(3);	choose_color->labelsize(14);
	choose_color->box(FL_RSHADOW_BOX);	choose_color->callback(do_color);		
	
	connected = new Fl_Check_Button(60,280,120,25,
									" Connect dots with line?");
	connected->labelfont(FL_HELVETICA_BOLD);	connected->labelsize(14);
	connected->value(line_draw);			connected->callback(connect_cb);	

	apply = new Fl_Button(125,320,155,30,"Display the curve");
	apply->color(62);		apply->labelcolor(FL_BLACK);
	apply->labelfont(3);	apply->labelsize(16);	
	apply->box(FL_RSHADOW_BOX);		apply->callback(doPLOT);

	//////////////////////////////////////////////////////////////////
	//////////////      S P E C T R O G R A M       //////////////////
	//////////////////////////////////////////////////////////////////
	
	Fl_Box* line1 = new Fl_Box(FL_FRAME_BOX,10,370,400,32,""); line1->color(6);
	Fl_Box* label7 = new Fl_Box(150,373,120,25,
								"Set Parameters for Spectrogram Calculation");
	label7->labelfont(FL_HELVETICA_BOLD_ITALIC); label7->labelsize(15);
	
	Fl_Box* label8 = new Fl_Box(50,413,100,50,"Resolution\nBalancing Factor:");
	label8->labelfont(FL_HELVETICA_BOLD); label8->labelsize(14);
	
	Fl_Box* label9 = new Fl_Box(180,390,200,50,
									"Increase Vertical Resolution @->");
	label9->labelfont(FL_HELVETICA_ITALIC); label9->labelsize(13);		
	
	balance = new Fl_Value_Slider(180, 425, 200, 25,
									"@<- Increase Horizontal Resolution");
	balance->type(FL_HOR_SLIDER);
	balance->bounds(0.1, 4.0);			// Set up bounds of the slider
	balance->value(k_res);				// Set up the default value
	balance->step(0.1);	
	balance->selection_color(FL_BLUE);
	balance->labelsize(13); balance->labelfont(FL_HELVETICA_ITALIC);
	balance->callback(slider_input);	// Set up the callback function
	balance->show();
	

	Fl_Box* label10 = new Fl_Box(50,478,100,50,
										 "Relative Taper Size:");
	label10->labelfont(FL_HELVETICA_BOLD); label10->labelsize(14);
	
	tapersz = new Fl_Value_Slider(180, 490, 200, 25,
								"% of the displayed depth range");
	tapersz->type(FL_HOR_SLIDER);
	tapersz->bounds(0.0, 20.0);			// Set up bounds of the slider
	tapersz->value(2.0);				// Set up the default value
	tapersz->step(1.0);	
	taperk = 2.0;
	tapersz->selection_color(FL_GREEN);
	tapersz->labelsize(13); tapersz->labelfont(FL_HELVETICA_ITALIC);
	tapersz->callback(slidek_input);	// Set up the callback function
	tapersz->show();	
	

	sprintf(cbuf2,"Window Size [%s]\nfor Detrending: ",unit[dep_unit]);
			  	
	Fl_Box* label11 = new Fl_Box(50,550,100,50,cbuf2);
	label11->labelfont(FL_HELVETICA_BOLD); label11->labelsize(13);
	
	Fl_Box* label12 = new Fl_Box(180,530,200,50,
							"Equivalent to high-pass filtering");
	label12->labelfont(FL_HELVETICA_ITALIC); label12->labelsize(13);		
	
	i = WIN/10; win_mean = 10*i;  // rounded half window size, samples
	demean = new Fl_Value_Slider(180,565,200,25);
	demean->type(FL_HOR_SLIDER);
	demean->bounds(0, win_mean*dS);	// Set up bounds of the slider
	demean->value(win_mean*dS);		// Set up the default value, length units
	win_mean /= 2;					// Half window size, samples
	demean->step(10);
	demean->selection_color(FL_RED);
	demean->labelsize(13); demean->labelfont(FL_HELVETICA_ITALIC);
	demean->callback(slides_input);	// Set up the callback function
	demean->show();
	
	
	Fl_Input_Choice in_pal(180,615,180,25,"Choose colormap: ");
	in_pal.callback(choice_pal, (void*)&in_pal);
	
	in_pal.add("  0 - Blue-White-Red");	
	in_pal.add("  1 - Yellow-Green-Blue");
	in_pal.add("  2 - Pink-Green");	
	in_pal.add("  3 - Grey-scale"); 
	
	in_pal.value(1);  // default palette #
	in_pal.labelsize(14);
	in_pal.labelfont(FL_HELVETICA_BOLD);
	in_pal.align(Fl_Align(FL_ALIGN_LEFT));
	
	
	palette = new Draw_pal(90,670,232,16);	// 10*23+2 pixels
	Fl_Box* pal2 = new Fl_Box(80,648,250,20,
					"@<- LOW     Spectral Magnitude     HIGH @->");
	pal2->labelfont(FL_HELVETICA_BOLD);	pal2->labelsize(12);	
	

	
	Fl_Box* label13 = new Fl_Box(50,695,100,50,"Gain Spectra:");
	label13->labelfont(FL_HELVETICA_BOLD); label13->labelsize(14);

	gainsp = new Fl_Value_Slider(180, 708, 200, 25);
	gainsp->type(FL_HOR_SLIDER);
	gainsp->bounds(0.1, 4.0);			// Set up bounds of the slider
	gainsp->value(gain);				// Set up the default value
	gainsp->step(0.1);	
	gainsp->selection_color(5);
	gainsp->labelsize(13); gainsp->labelfont(FL_HELVETICA_ITALIC);
	gainsp->callback(slideg_input);	// Set up the callback function
	gainsp->show();		
	
	
					
	specgram = new Fl_Button(150,760,125,30,"Spectrogram");
	specgram->color(62);				specgram->labelcolor(FL_BLACK);
	specgram->labelfont(3);				specgram->labelsize(16);	
	specgram->box(FL_RSHADOW_BOX);		specgram->callback(doSPEC);		
	
	/////////////////////////////////////////////////////////////////
	
	end_this = new Fl_Button(165,810, 80, 30, "Exit");
	end_this->color(208);			end_this->labelcolor(FL_BLACK);
	end_this->labelfont(3);			end_this->labelsize(18);	
	end_this->box(FL_RSHADOW_BOX);	end_this->callback(exit_b);	

	opts_window->end(); 
	opts_window->show();
	/////////////////////////////////////////////////////////////////
	return(Fl::run());
   return 0;
}
//////////////////////////////////////////////////////////////////////
/****    END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN    ****/
//////////////////////////////////////////////////////////////////////
/****               Called Function Definition                   ****/
//////////////////////////////////////////////////////////////////////
/**
	==============    LAS file format standards    ==================
				Mandatory sections that make up an LAS file:
	~V  - contains version and wrap mode information as following:
 VERS.                2.00:   CWLS log ASCII Standard -VERSION 2.00
 WRAP.						YES:  Multiple lnes
				or
 WRAP.                  NO:   One line per depth step 
 Non-wrapped line in the data section is no longer than 256 characters  
 
	~W  - contains well identification with mnemonics as indicated:
 STRT.FT                                        0.0000 : Starting Depth
 STOP.FT                                    12820.0000 : Ending Depth
 STEP.FT                                        0.5000 : Level Spacing
 NULL.                                       -999.2500 : Absent value
		( units of depth allowed are FT or M ) 	
	
	~C  - contains curve information
	~P  - contains parameters or constants  (OPTIONAL secion)
	~O  - contains other information			 (OPTIONAL secion)
	~A  - contains ASCII log data with depths in the first column
	
FLAGS are used to assist computers in identifying specific lines in a LAS file:

"~" (tilde) is used to mark the beginning of a section in a file.
It is the first non-space and non-quotation character on a line

"#" is used to indicate a comment line. Can appear only above the data section.

"." and ":" in sections other than the data secton are used to delimit 
information within the line. Info to the right of the colon is a detailed
definition of the mnemonic located to the left of the colon. The dot is used
to separate two mnemonics. Spaces may occur to the right or left of "." or ":".
**/
void read_file_info(const char *name) // Reads ALL log-data into logs[]  
{
	int i,j,k,n,pass,counter;
	char *ptr, buf[80], bigbuf[801], keys[]="-1234567890";
	float val;
	double dval,wl;

	if ((in_fp = fopen(name, "r")) == NULL)
	{fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}

	///////////////////////////////////////////////////////////////
	/////    Read LAS sections with mandatory information		//////
	///////////////////////////////////////////////////////////////	
	pass = 0;	fseek(in_fp, 0L, SEEK_SET);	// go to beginning
	while(fgets(bigbuf,800,in_fp))			// while != NULL
	{
		if(strstr(bigbuf,"WRAP") != NULL) // located a string with "WRAP." 
		{
			pass = 1;
			if(strstr(bigbuf,"YES") != NULL) // "YES" in the same string
			{ fl_message("Text wrapping is not supported in this program!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }
			else break;
		}
	}
	if(pass==0)
	{ fl_message("1. LAS mandatory information was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }


	pass = 0;	fseek(in_fp, 0L, SEEK_SET);	// go to beginning
	while(fgets(bigbuf,800,in_fp))			// while != NULL
	{
		if(strstr(bigbuf,"STRT") != NULL) // located a string with "STRT" 
		{
			j=strcspn(bigbuf,".");  dep_unit = 2; // unknown by default
			strncpy(buf,bigbuf+j,2); buf[2]='\0';
			if(buf[1]=='M') dep_unit=0;
			if(buf[1]=='F') dep_unit=1;
		
			i=strcspn(bigbuf,keys); // search for the first digit occurrence
			j=strcspn(bigbuf,":");  // search for the first colon occurrence
			strncpy(buf,bigbuf+i,j-i);
			buf[j]='\0';				// terminating NULL
			sscanf(buf,"%f",&depth_sta);
			
			fgets(bigbuf,800,in_fp);
			if(strstr(bigbuf,"STOP") != NULL) 
			{
				i=strcspn(bigbuf,keys);
				j=strcspn(bigbuf,":");
				strncpy(buf,bigbuf+i,j-i);
				buf[j]='\0';
				sscanf(buf,"%f",&depth_end);			
			}
			else break;
			
			fgets(bigbuf,800,in_fp);					
			if(strstr(bigbuf,"STEP") != NULL) 
			{
				i=strcspn(bigbuf,keys);
				j=strcspn(bigbuf,":");
				strncpy(buf,bigbuf+i,j-i);
				buf[j]='\0';
				sscanf(buf,"%f",&dS);			
			}
			else break;
			
			fgets(bigbuf,800,in_fp);					
			if(strstr(bigbuf,"NULL") != NULL) 
			{
				i=strcspn(bigbuf,keys);
				j=strcspn(bigbuf,":");
				strncpy(buf,bigbuf+i,j-i);
				buf[j]='\0';
				sscanf(buf,"%f",&no_value);			
			}
			else break;			
			
			pass=1; break;			
		}
	}
	if(pass==0)
	{ fl_message("2. LAS mandatory information was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }

	///////////////////////////////////////////////////////////////
	/////    Get Info on the Number-of-Curves and Samples 	//////
	///////////////////////////////////////////////////////////////
	pass = 0;	fseek(in_fp, 0L, SEEK_SET);	// go to beginning
	while(fgets(bigbuf,800,in_fp))				// while != NULL
	{
		if(strstr(bigbuf,"~C") != NULL) // located Curve Info Section 
		{
			k=0;
			for(i=0;i<MAX_CURVE;i++)
			{
				fgets(bigbuf,800,in_fp); 
				if(bigbuf[0]=='#') continue;
				if(bigbuf[0]=='~') break;
				
				sprintf(buf,"%3d",k);
				strcat(buf,"  ");  
				strncat(buf,bigbuf,17);
				j=strcspn(buf,"/");    buf[j]='_';	// FLTK doesn,t like "/"
				buf[21]='\0';						
				strncpy(curve[k],buf,22);				
				
				k++; if(k>=MAX_CURVE)
				{ fl_message("ERROR: excessive number of curves\n\
Close the window to exit"); fclose(in_fp); exit(1); }				
			}
			pass=1; Ncur=k; break;		// Number of curves (logs) in file
		}
	}
	if(pass==0)
	{ fl_message("3. LAS mandatory information was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }


	pass=i=0;	fseek(in_fp, 0L, SEEK_SET);	// go to beginning
	while(fgets(bigbuf,800,in_fp))			// while != NULL
	{
		if(pass==1) i++;	
		if(strstr(bigbuf,"~A") != NULL) pass=1; // located Data Section
	}	
	if(pass==0)
	{ fl_message("4. LAS data section  was not found!\n\
Close the window to exit..."); fclose(in_fp); exit(1); }
	
	Ndat = i;		// Number of data samples in each curve (log)
	
	mini = new (nothrow) float[Ncur];			// for logs scaling
	maxi = new (nothrow) float[Ncur];			// for logs scaling
	avri = new (nothrow) float[Ncur];			// for logs mean	
	logs = new (nothrow) float[Ncur * Ndat]; 	// Allocate MEMORY  ~10 MB
	if(logs == 0)
	{ fl_message("ERROR: NO memory to accomodate log data!\n\
Close the window to exit...");	fclose(in_fp);  exit(1); }

	///////////////////////////////////////////////////////////////
	///////    Position File pointer on a Data Section	  /////////
	fseek(in_fp, 0L, SEEK_SET);
	while(fgets(bigbuf,800,in_fp)) if(strstr(bigbuf,"~A") != NULL) break;
	///////////////////////////////////////////////////////////////
	///////    Read Data Section to the logs[..] array	  /////////
	///////////////////////////////////////////////////////////////			
	i=0;							// i = data sample index
	while(fgets(bigbuf,800,in_fp))
	{
		if(i>=Ndat) 
		{ fl_message("5. Corrupted Log-Data Block\n\
Close the window to exit"); fclose(in_fp); exit(1); }
		ptr = strtok (bigbuf," ");  
		j=0;						// j = curve number index
		while (ptr != NULL)
		{
			sscanf(ptr,"%f",logs+Ncur*i+j);
			ptr = strtok (NULL, " ");	j++;			
		}
		if(j != Ncur) 
		{ fl_message("6. Corrupted Log-Data Block\n\
Close the window to exit"); fclose(in_fp); exit(1); }
		i++;
	}
	//  Final data test...
	//  Most 'Corrupted Log-Data Block' messages are due to strange symbols
	//  (like ?) found at the end-of-file with 'Xcode' editor - delete them!	
	i=(int)logs[0];  k=(int)logs[Ncur*(Ndat-1)];
	j=(int)depth_sta;  n=(int)depth_end;
	if(i == j && k == n) ;
	else { fl_message("7. Corrupted Log-Data Block\n\
Close the window to exit"); fclose(in_fp); exit(1); }

	fclose(in_fp);
	
	///////////////////////////////////////////////////////////////
	////	Estimate the Min-Max and Avr. of log data		   ////
	///////////////////////////////////////////////////////////////		
	for(j=0;j<Ncur;j++)				// Search for Min-Max values 
	{
		mini[j]=FLT_MAX;  maxi[j]=FLT_MIN; dval=0.0; counter=0;
		for(i=0;i<Ndat;i++)
		{ 
			val=logs[Ncur*i+j];
			if(val>no_value) 
			{ 
				dval += val; counter++;
				if(val>maxi[j]) maxi[j]=val; 
				if(val<mini[j]) mini[j]=val; 
			}
		}
		
		avri[j] = (float)(dval/counter);
		val = (maxi[j]-mini[j]);  // Round-up  Min-Max vales
		
		if(val>1000.0)				
		{ 
			i=(int)(avri[j]/100.0); avri[j]=100.0*i;		
			i=(int)(mini[j]/100.0); mini[j]=100.0*i; 
			i=(int)(maxi[j]/100.0); maxi[j]=100.0*(i+1);  continue;
		}
		if(val>100.0) 
		{ 
			i=(int)(avri[j]/10.0); avri[j]=10.0*i;	
			i=(int)(mini[j]/10.0); mini[j]=10.0*i; 
			i=(int)(maxi[j]/10.0); maxi[j]=10.0*(i+1);  continue;
		}
		if(val>10.0) 
		{ 
			i=(int)avri[j]; avri[j]=i;		
			i=(int)mini[j]; mini[j]=i; 
			i=(int)maxi[j]; maxi[j]=(i+1); continue;
		}
		if(val>1.0) 
		{
			i=(int)(avri[j]*10.0); avri[j]=0.1*i;		 
			i=(int)(mini[j]*10.0); mini[j]=0.1*i; 
			i=(int)(maxi[j]*10.0); maxi[j]=0.1*(i+1); continue;
		}						
		if(val>0.1) 
		{
			i=(int)(avri[j]*100.0); avri[j]=0.01*i;		 
			i=(int)(mini[j]*100.0); mini[j]=0.01*i; 
			i=(int)(maxi[j]*100.0); maxi[j]=0.01*(i+1); continue;
		}
	}	

	dS = (depth_end-depth_sta)/(Ndat-1);  // More precise..?
	dep_sta=depth_sta;  dep_end=depth_end;	// default plotting range
 	min_val=mini[curve_no];  max_val=maxi[curve_no];
	
	pal = YGB_pal;							// Default color palette

	///////////////////////////////////////////////////////////////
	/////////////////      SPECTROGRAM      ///////////////////////

	//  WIN - number of data samples corresponding 150 meters (492 ft)
	if(dep_unit==1) WIN = (int)((3.28*DWIN)/dS);	// feet
	else WIN = (int)(DWIN/dS);	// meters and unknown units

	// Minimum wavelength should be more than 2 samples
	// Maximum wavelength [in samples] should not exceed WIN samples	
	wls_min = 2;		// Min. (Nyquist) wavelength in samples
	wls_max = WIN;		// Maximum wavelength in samples

	//////     Array of Gaussian [bell-shaped]  Weights	 	////////
	for(j=0;j<FOLD;j++) bell[j] =(double*)malloc(WIN*sizeof(double));

//////////////////////////////////////////////////////////////////
}
////////////////////////////////////////////////////////////////
//	     Cubic Convolution Interpolation [R. Keys, 1981].
//  Four data values from an array  Y[0..NS) measured at X values
//  from 1.0 to NS with increment = 1.0 are used every time
//  to RETURN an interpolated value at some X = X_intr
////////////////////////////////////////////////////////////////
float CCI(float *Y, int NS, float X_intr)
{
	int i;
	float X,y_1,y0,y1,y2;
	double dx,dx2,dx3,k_1,k0,k1,k2;
	
	if(X_intr<=1.0) return Y[0];
	if(X_intr>=(float)NS) return Y[NS-1];
	
	for(i=0;i<NS;i++) 
	{  X = (float)(i+1);   if(X>X_intr) break; }
	X-=1.0;  i--;
	
	y0=Y[i]; 	
	if(i<1) y_1=y0; else y_1=Y[i-1];
	if(i>NS-2) y1=y0; else y1=Y[i+1]; 
	if(i>NS-3) y2=y1; else y2=Y[i+2];
	
	dx = (double)(X_intr - X);
	dx2=dx*dx;  dx3=dx*dx2;		
	
	k_1 = (2.0*dx2 - dx3 - dx)/2.0;
	k0 = (3.0*dx3 - 5.0*dx2 + 2.0)/2.0;
	k1 = (4.0*dx2 - 3.0*dx3 + dx)/2.0;
	k2 = (dx3 - dx2)/2.0;
	
	return (k_1*y_1 + k0*y0 + k1*y1 + k2*y2);	
}
//
/////  Caclulates a set of Gaussian bell-shaped curves with
/////  variable width that depends on the resolution factor
//
//	Bell-shaped Gaussian = a*exp(-(x-b)*(x-b)/(2*wl*wl));
// a = height of the bell's peak, b = position of the center
// of the peak, and wl = width of the 'bell' (standard deviation)
/////////////////////////////////////////////////////////////////
void bells(int win_sz, double res_factor)
{                       // 0.1 < k_res < 4.0 
	int i,j;
	float val;
	double wl,dval;
	
	for(j=0;j<FOLD;j++)
	{ // computes 2^x
		wl = exp2(res_factor*j+res_factor*res_factor+1.5);
		norm[j]=0.0;
		for(i=0;i<win_sz;i++)   // win_sz - window size [samples]
		{	// considering zero mean
			val = (float)(win_sz/2-i)*(win_sz/2-i);
			dval = exp(-1.0*val/(2.0*wl*wl)); // e^x
			bell[j][i] = dval;		norm[j] += dval;
		}	
	}
}
////////////////////////////////////////////////////////////////
void pal_interp(float palette[], int pal_size, float x, short colors[])
{  // based on "index_search" routine for linear interpolation
	int i,ilo,ihi;
	float dx,dr,dg,db,r0,g0,b0,rf,gf,bf;	
	
	ilo = 0;
	ihi = pal_size-1;
	
	while(ihi > ilo + 1) 
	{
		i = (ihi + ilo)/2;
		if(palette[i] > x) ihi = i;
		else ilo = i;
	}
	
	dx = (x-palette[ilo])/(palette[ilo+1]-palette[ilo]);
	r0 = palette[pal_size+ilo];
	g0 = palette[pal_size*2+ilo];
	b0 = palette[pal_size*3+ilo];
	dr = palette[pal_size+ilo+1] - r0;
	dg = palette[pal_size*2+ilo+1] - g0;
	db = palette[pal_size*3+ilo+1] - b0;
	
	rf = r0 + dx*dr; if(rf<0.1) rf =0.1; if(rf>255.1) rf =255.1;
	gf = g0 + dx*dg; if(gf<0.1) gf =0.1; if(gf>255.1) gf =255.1;
	bf = b0 + dx*db; if(bf<0.1) bf =0.1; if(bf>255.1) bf =255.1;
	
	colors[0] = (short)rf;
	colors[1] = (short)gf;	
	colors[2] = (short)bf;	
}
